use std::{
    collections::{HashMap, HashSet},
    time::Duration,
};

use futures_lite::StreamExt;
use lapin::{
    options::{BasicAckOptions, QueueDeclareOptions},
    protocol::basic::AMQPProperties,
    types::{AMQPValue, FieldArray, FieldTable, LongLongInt, ShortString},
    Channel, Connection, Consumer,
};

use rand::Rng;
use tokio::time::sleep;

// static STREAM_NAME: &'static str = "rustic-events";
static STREAM_NAME: &'static str = "dynamic-test";

static MESSAGES: &[&str] = &[
    "0::abc", "0::hello", "1::1", // ignored
    "0:sub:1", "1::2", "1:sub:2", "2::1", "3::1",  // ignored
    "delay", // ignored
    "2:sub:3", "3::1", "3::2", "3::3", "3::4", "3::5", "3::6", "delay", // ignored
    "3::7", "3::8", "3:sub:4", "4:sub:5", "5:sub:6", "6:sub:7", "7:sub:8", "8:sub:9",
];

struct Worker {
    channel: Channel,
    topics: HashSet<String>,
    offset: Option<LongLongInt>,

    // testing
    received: HashSet<String>,
}

impl Worker {
    async fn from(channel: Channel) -> Self {
        let mut args: FieldTable = Default::default();
        args.insert(
            "x-queue-type".into(),
            AMQPValue::LongString("stream".into()),
        );
        args.insert(
            "x-max-length-bytes".into(),
            AMQPValue::LongLongInt(5_000_000_000),
        );

        channel
            .queue_declare(
                STREAM_NAME,
                QueueDeclareOptions {
                    durable: true,
                    ..Default::default()
                },
                args,
            )
            .await
            .unwrap();

        channel.basic_qos(100, Default::default()).await.unwrap();

        Worker {
            channel,
            topics: Default::default(),
            offset: Default::default(),
            received: Default::default(),
        }
    }

    fn sub(&mut self, topic: String) {
        self.topics.insert(topic);
    }

    async fn read_to_end(mut consumer: Consumer) {
        while let Some(delivery) = consumer.next().await {
            let delivery = delivery.expect("error in consumer");
            delivery.ack(BasicAckOptions::default()).await.expect("ack");
        }
    }

    async fn worker(&mut self) {
        println!("Starting worker...");

        let mut args: FieldTable = Default::default();
        let mut filter: FieldArray = Default::default();
        for topic in &self.topics {
            filter.push(AMQPValue::LongString(topic.as_str().into()));
        }

        args.insert("x-stream-filter".into(), AMQPValue::FieldArray(filter));

        if let Some(offset) = self.offset {
            args.insert("x-stream-offset".into(), AMQPValue::LongLongInt(offset));
        }

        let tag: String = rand::thread_rng()
            .sample_iter::<char, _>(rand::distributions::Standard)
            .take(32)
            .collect();

        let mut consumer = self
            .channel
            .basic_consume(STREAM_NAME, &tag, Default::default(), args)
            .await
            .unwrap();

        println!("Initialised a new consumer...");

        while let Some(delivery) = consumer.next().await {
            let mut break_out = false;

            let delivery = delivery.expect("error in consumer");
            delivery.ack(BasicAckOptions::default()).await.expect("ack");
            // println!("Delivery! {:?}", delivery);

            let headers: HashMap<String, AMQPValue> = delivery
                .properties
                .headers()
                .as_ref()
                .map(|table| {
                    table
                        .into_iter()
                        .map(|(k, v)| (k.to_string(), v.clone()))
                        .collect()
                })
                .unwrap_or_default();

            let filter_value = headers
                .get("x-stream-filter-value")
                .expect("`x-stream-filter-value` not present in message!");

            if self
                .topics
                .contains(&filter_value.as_long_string().unwrap().to_string())
            {
                let data = std::str::from_utf8(&delivery.data).unwrap();
                println!("Received: {}", data);
                self.received.insert(data.to_string());
                println!(
                    "Received {} out of {}",
                    self.received.len(),
                    MESSAGES.len() - 4
                );

                let mut args = data.split(':');
                args.next();

                match args.next() {
                    Some("sub") => {
                        self.sub(args.next().unwrap().to_string());
                        break_out = true;
                    }
                    _ => {}
                }
            }

            let stream_offset = headers
                .get("x-stream-offset")
                .expect("`x-stream-offset` not present in message!");

            self.offset = Some(stream_offset.as_long_long_int().unwrap() + 1);

            if break_out {
                break;
            }
        }

        if let Err(err) = self.channel.basic_cancel(&tag, Default::default()).await {
            eprintln!("Failed to close consumer! {:?}", err);
        } else {
            println!("Closed current consumer, reading it to the end...");
            Worker::read_to_end(consumer).await;
        }
    }

    async fn run(mut self) {
        self.sub("0".to_string());
        loop {
            self.worker().await
        }
    }
}

#[tokio::main]
async fn main() {
    let addr = std::env::var("AMQP_ADDR").unwrap_or_else(|_| "amqp://127.0.0.1:5672/%2f".into());
    let conn = Connection::connect(&addr, Default::default())
        .await
        .unwrap();

    let channel = conn.create_channel().await.unwrap();
    tokio::spawn(async move {
        sleep(Duration::from_secs(1)).await;

        // same as worker
        let mut args: FieldTable = Default::default();
        args.insert(
            "x-queue-type".into(),
            AMQPValue::LongString("stream".into()),
        );
        args.insert(
            "x-max-length-bytes".into(),
            AMQPValue::LongLongInt(5_000_000_000),
        );

        channel
            .queue_declare(
                STREAM_NAME,
                QueueDeclareOptions {
                    durable: true,
                    ..Default::default()
                },
                args,
            )
            .await
            .unwrap();
        // end code

        for msg in MESSAGES {
            if msg == &"delay" {
                sleep(Duration::from_secs(1)).await;
                continue;
            }

            let mut args = msg.split(':');

            let mut headers: FieldTable = Default::default();
            headers.insert(
                "x-stream-filter-value".into(),
                AMQPValue::LongString(args.next().unwrap().into()),
            );

            let properties: AMQPProperties = Default::default();

            channel
                .basic_publish(
                    "",
                    STREAM_NAME,
                    Default::default(),
                    msg.as_bytes(),
                    properties.with_headers(headers),
                )
                .await
                .unwrap();
        }
    });

    let worker = Worker::from(conn.create_channel().await.unwrap()).await;
    worker.run().await
}
