use std::{
    collections::{hash_set::Iter, HashSet},
    time::Duration,
};

use futures_lite::StreamExt;
use lapin::{
    options::{BasicAckOptions, QueueDeclareOptions},
    protocol::basic::AMQPProperties,
    types::{AMQPValue, FieldArray, FieldTable, LongLongInt, ShortString},
    Channel, Connection, Consumer,
};

use rand::Rng;
use tokio::time::sleep;

#[derive(Debug)]
enum Error {
    SomethingHappened,
}

type Result<T> = std::result::Result<T, Error>;

struct Stream {
    channel: Channel,
    name: String,
}

impl Stream {
    async fn new(conn: &Connection, name: String) -> Result<Self> {
        let channel = conn
            .create_channel()
            .await
            .map_err(|_| Error::SomethingHappened)?;

        let mut args: FieldTable = Default::default();
        args.insert(
            "x-queue-type".into(),
            AMQPValue::LongString("stream".into()),
        );
        args.insert(
            "x-max-length-bytes".into(),
            AMQPValue::LongLongInt(5_000_000_000),
        );

        channel
            .queue_declare(
                &name,
                QueueDeclareOptions {
                    durable: true,
                    ..Default::default()
                },
                args,
            )
            .await
            .unwrap();

        channel.basic_qos(100, Default::default()).await.unwrap();

        Ok(Stream { channel, name })
    }
}

trait MessageHandler {
    async fn handle(&mut self, topics: &mut Topics) -> Result<()>;
}

#[derive(Default)]
struct Topics {
    filter: HashSet<String>,
    modified: bool,
}

impl Topics {
    pub fn iter(&mut self) -> Iter<String> {
        self.filter.iter()
    }

    pub fn add(&mut self, topic: String) {
        unimplemented!()
    }

    pub fn remove(&mut self, topic: &str) {
        unimplemented!()
    }
}

struct ListenWorker {
    stream: Stream,
    pub topics: Topics,
    offset: Option<i64>,
}

impl ListenWorker {
    /// Create a new listen worker
    async fn new(conn: &Connection, stream_name: String) -> Result<Self> {
        Ok(ListenWorker {
            stream: Stream::new(conn, stream_name).await?,
            topics: Default::default(),
            offset: None,
        })
    }

    /// Read a consumer to the end and acknowledge any still pending messages
    async fn read_to_end(mut consumer: Consumer) {
        while let Some(delivery) = consumer.next().await {
            let delivery = delivery.expect("error in consumer");
            delivery.ack(BasicAckOptions::default()).await.expect("ack");
        }
    }

    /// Establish a new consumer for the stream and start reading messages
    async fn consume(&mut self) {
        let mut args: FieldTable = Default::default();
        let mut filter: FieldArray = Default::default();
        for topic in &self.topics {
            filter.push(AMQPValue::LongString(topic.as_str().into()));
        }
    }

    /// Start the worker
    pub async fn start(&mut self) -> ! {
        loop {
            self.consume().await
        }
    }

    /// Signal the worker to stop and flush acknowledgements
    pub async fn stop() {
        unimplemented!()
    }
}

#[tokio::main]
async fn main() {
    let addr = std::env::var("AMQP_ADDR").unwrap_or_else(|_| "amqp://127.0.0.1:5672/%2f".into());
    let conn = Connection::connect(&addr, Default::default())
        .await
        .unwrap();

    struct HandleMsg {}
    impl MessageHandler for HandleMsg {
        async fn handle(&mut self, topics: &mut Topics) -> Result<()> {
            // topics.add("hello".to_string());

            Ok(())
        }
    }

    let mut worker = ListenWorker::new(&conn, "stream".to_string())
        .await
        .unwrap();

    worker.topics.add("user id".to_string());
    worker.start().await
}
